@@ -28,6 +28,17 @@ npchar_to_typ = {'b': 'char',
 typ_to_npchar = dict( (v,k) for k,v in npchar_to_typ.items() )
 
 
+nptyp_to_cl = {np.dtype('float16'): 'half',
+                np.dtype('float32'): 'float',
+                np.dtype('float64'): 'double',
+                np.dtype('uint8'): 'uchar',
+                np.dtype('int16'): 'short',
+                np.dtype('int32'): 'int',
+                np.dtype('int64'): 'long'}
+cltyp_to_np = dict( (v,k) for k,v in nptyp_to_cl.items() )
+
+
+
 def verify_apply( func, argtypes ):
     """
     verify_apply( func, argtypes )
@@ -35,7 +46,8 @@ def verify_apply( func, argtypes ):
     verify that function accepts arg types given
     returns return-type of function
     """
-    #if type(func) == np.ufunc:
+
+    #argtypes = [nptyp_to_cl[t] for t in argtypes]
     matching_ret = None
     for t in func.types:
         args, ret = t.split('->')
@@ -61,7 +73,7 @@ def special_funcs( modname, funcname, symbol_lookup, args ):
         # FIXME: should we check the type of args?  also, need we worry about the return type required
         return 'convert_int_rtz', 'int'
     if not modname and funcname == 'float':
-        return 'convert_float_rtz', 'float'
+        return 'convert_double', 'double'
 
     # FIXME: enforce args
     import importlib
@@ -80,6 +92,9 @@ def special_funcs( modname, funcname, symbol_lookup, args ):
 
 
 def conv( el, symbol_lookup, declarations=None ):
+    """
+    returns <openCL-C string representation>, <openCL type>
+    """
     def is_float(s):
         if s.startswith('float'):
             return s
@@ -97,7 +112,7 @@ def conv( el, symbol_lookup, declarations=None ):
     def cpow( left_el, right_el ):
         (lval, ltyp), (rval, rtyp) =  (_conv(left_el), _conv(right_el))
         assert None in (ltyp,rtyp) or ltyp == rtyp, "pow requires types match; got %s, %s" % (ltyp,rtyp)
-	return "pow( %s, %s )" % (lval,rval), ltyp if ltyp is not None else rtyp
+	return "pow( %s, %s )" % (lval,rval), ltyp or rtyp
 
     def cnumeric( s ):
 	try:
@@ -114,6 +129,7 @@ def conv( el, symbol_lookup, declarations=None ):
             raise ValueError("comparitor not supported: '%s'" % str(s))
 
     name = el.get('_name')
+
     if name == 'Name':
 	# identity function
 	iden = el.get('id')
@@ -136,10 +152,10 @@ def conv( el, symbol_lookup, declarations=None ):
 	operand, typ = _conv( operand )
 	# Invert | Not | UAdd | USub
 	[op] = el.findall('./op')
-	return {'Invert':	('~' + operand, typ),
-		'Not':		('!' + operand, typ),
-		'UAdd':		(operand, typ),
-		'USub':		('-' + operand, typ) }[ op.get('_name') ]
+	return {'Invert':	'~' + operand,
+		'Not':		'!' + operand,
+		'UAdd':		operand,
+		'USub':		'-' + operand }[ op.get('_name') ], typ
 
     if name == 'BinOp':
 	[op] = el.findall('./op')
@@ -254,6 +270,9 @@ def pprint( s ):
 
 
 def lambda_to_kernel( lmb, types, bindings=None ):
+    """
+    @types -- numpy types
+    """
     # lstrip, b/c there's likely whitespace that WILL get parsed
     src = ast.parse( inspect.getsource( lmb ).lstrip() )
     root = ET.fromstring( ast2xml.ast2xml().convert(src) )
@@ -272,7 +291,7 @@ def lambda_to_kernel( lmb, types, bindings=None ):
         target_name = None
         # returns: requires_declaration, type, string_representation
         if s in declarations:
-            return True, declarations[s], (s + '[gid]')  # requires_declaration=True, b/c it's moot
+            return True, nptyp_to_cl[declarations[s]], (s + '[gid]')  # requires_declaration=True, b/c it's moot
         else:
             # hackiness here:
             target_name = re.match( r'(\w+)\[?', s ).group(1)
@@ -283,8 +302,9 @@ def lambda_to_kernel( lmb, types, bindings=None ):
 
     [body] = func.findall("./body")
     kernel_body, result_typ = conv(body, symbol_lookup=symbol_lookup, declarations=declarations)
+    numpy_typ = cltyp_to_np[result_typ]
 
-    sigs = ["__global const %s *%s" % (typ,aname) for typ,aname in zip(types,argnames)] \
+    sigs = ["__global const %s *%s" % (nptyp_to_cl[ typ ], aname) for typ,aname in zip(types,argnames)] \
            if types \
              else ["__global const float *%s" % aname for aname in argnames]
 
@@ -304,7 +324,7 @@ __kernel void sum( %(sigs)s ) {
     if bindings is None:
         kernel = "/* NOTE: without numpy bindings, some types might be incorrectly annotated as None */" + kernel
 
-    return (argnames, kernel)
+    return (argnames, kernel, result_typ)
 
 
 
@@ -312,7 +332,6 @@ def function_to_kernel( f, types, bindings=None ):
     #####
     # not a lambda, but a traditional function
     # lstrip, b/c there's likely whitespace that WILL get parsed
-
     src = ast.parse( inspect.getsource( f ).lstrip() )
     root = ET.fromstring( ast2xml.ast2xml().convert(src) )
 
@@ -323,7 +342,7 @@ def function_to_kernel( f, types, bindings=None ):
     idx_name = argnames.pop(0)
     results_name = argnames.pop(0)
 
-    argname_to_type = dict( zip( argnames, types ) ) if types \
+    argname_to_type = dict( (nom, nptyp_to_cl[ntyp]) for nom, ntyp in zip(argnames, types ) ) if types \
                       else dict( (a, None) for a in argnames )
 
     declarations = {}
@@ -353,18 +372,18 @@ def function_to_kernel( f, types, bindings=None ):
     #assignments = [conv(el, symbol_lookup=symbol_lookup) for el in func.findall("./body/_list_element[@_name='Assign']")]
 
     [body] = func.findall("./body")
-    kernel_body, typ = conv(body, symbol_lookup=symbol_lookup, declarations=declarations)
+    _, typ = conv(body, symbol_lookup=symbol_lookup, declarations=declarations)
 
     # res_g should appear in declarations, as we want to know its inferred type, but we don't actually want to declare it
     result_typ = declarations['res_g']
     del declarations['res_g']
 
-    sigs = ["__global const %s *%s" % (typ,aname) for typ,aname in zip(types,argnames)] \
+    sigs = ["__global const %s *%s" % ( nptyp_to_cl[ntyp], aname) for ntyp,aname in zip(types,argnames)] \
            if types else ["__global const float *%s" % aname for aname in argnames]
     sigs.append( '__global %s *res_g' % result_typ)
 
     input_sig =  ', '.join(sigs)
-    decl = '\n'.join( '%s %s;' % (typ,nom) for nom, typ in declarations.items())
+    decl = '\n'.join( '%s %s;' % (typ, nom) for nom, typ in declarations.items())
 
     kernel = """
 
@@ -380,7 +399,7 @@ __kernel void sum( %(sig)s ) {
     if bindings is None:
         kernel = "/* NOTE: without numpy bindings, some types might be incorrectly annotated as None */" + kernel
 
-    return (argnames, kernel)
+    return (argnames, kernel, result_typ)
 
 
 
